import decimal

def _transform_node(node_data, format_from, format_to):
    """ Helper to transform node elements from one format to another, recursive """
    if type(node_data) == str:
        return node_data
    if type(node_data) == format_from:
        node_data = round(format_to(node_data), 9) # 9 decimals always
    if type(node_data) == dict:
        for key in node_data.keys():
            if type(node_data[key]) == format_from:
                node_data[key] = round(format_to(node_data[key]), 9) # 9 decimals always
            elif type(node_data[key]) == dict or type(node_data[key]) == list:
                node_data[key] = _transform_data_types(
                    node_data[key], format_from, format_to)
    return node_data


def _transform_data_types(data, format_from, format_to):
    """ Helper for transforming data type """
    if type(data) == list:
        data_total = []
        for ele in data:
            if type(ele) == list:
                data_total.append(_transform_data_types(ele, format_from, format_to))
            else:
                data_total.append(_transform_node(ele, format_from, format_to))
    else:
        data_total = _transform_node(data, format_from, format_to)
    return data_total


def json_decimal_to_float(data):
    """ convert all decimal.Decimal numbers in the json root to float
    """
    return _transform_data_types(data, format_from=decimal.Decimal, format_to=float)


def json_float_to_decimal(data):
    """ convert all float numbers in the json root to decimal.Decimal
    """
    return _transform_data_types(data, format_from=float, format_to=decimal.Decimal)



from __future__ import absolute_import

import inspect
from marshmallow import Schema, fields
from lantern_sl.utils.dynamo import unique_id
from lantern_sl.utils.time import current_ts
from lantern_sl.utils.json import json_decimal_to_float
from lantern_sl.utils.request import http_response, http_error
from lantern_sl.serverless.base import ServerlessBase
from lantern_sl.controllers.dynamo_controller import (DynamoController, ExceptionInvalidDynamoControllerNotFound)

class ExceptionParamsError(Exception):
	pass

class ExceptionSerializerValidator(Exception):
	pass

FILTER_FIELDS = ["key", "value", "order_by", "limit", "next", "index", "order_type", "filter_expression", "invert_result"]
class FilterScheme(Schema):
	key = fields.Str(required=True)
	value = fields.Str(required=True)
	order_by= fields.Str()
	limit = fields.Number()
	next = fields.Str()
	index = fields.Str()
	order_type = fields.Str()
	filter_expression = fields.Raw()
	invert_result = fields.Boolean()


class ServerlessCrudBase(ServerlessBase):
	serializer = None # required to be implemented
	controller = None
	auto_add_id = True

	def __init__(self, event, context, table_name, debug=False):
		if table_name:
			self.table_name = table_name
		elif "DYNAMODB_TABLE" in context:
			self.table_name = context["DYNAMODB_TABLE"]
		self.json_response = self._get_json_response(event)
		self.controller = DynamoController(table_name=self.table_name, debug=debug)
		return super(ServerlessCrudBase, self).__init__(event, context)
	
	def verify_serializer_defined(self):
		if not self.serializer:
			raise Exception("serializer property is required! (property or parameter)")
		if not inspect.isclass(self.serializer):
			raise Exception("Serializer should be a class extending marshmallow.Schema")
		if not issubclass(self.serializer, Schema):
			raise Exception("Serializer should be a class extending marshmallow.Schema")
		return True
	
	def verify_data_serialized(self, data, many=False, partial=None):
		_serializer = self.serializer()
		errors = _serializer.validate(data, many=many, partial=partial)
		if errors:
			raise Exception("Serializer Errors: {}".format(errors))
		return True
		
	def _get_json_response(self, event):
		""" json_response in root event or inside 'crud' node """
		crud_node = event.get("crud", None)
		if crud_node:
			json_response = crud_node.get("json_response", None)
		else:
			json_response = event.get("json_response", False)
		return json_response
	
	def get_primary_keys(self, use_event_if_none=True, error_on_empty=False):
		""" find the primary node in the root node or inside 'crude' node """
		node = self.event
		AVAILABLE_NODES = ["crud", "query", "body"]
		for n in AVAILABLE_NODES:
			if node.get(n,None):
				node = node.get(n)
				break
		primary_keys = None
		if not primary_keys:
			primary_keys = node.get("primary_keys",None)
		if not primary_keys and "method" in self.event and self.event["method"].upper() in ["DELETE", "GET"]: # params by url
			primary_keys = node
		if not primary_keys and use_event_if_none:
			return self.event
		if not primary_keys and error_on_empty:
			raise ExceptionParamsError("'primary_keys' not found in event root neither in 'crud' node")
		return primary_keys
	
	def get_data(self, use_event_if_none=True, error_on_empty=False):
		""" find data data node in root or crud """
		""" find data data node in root or crud """
		# GET
		http_method = self.event.get('method', None)
		crud_node = self.event.get("crud", None)
		data_node = self.event.get("data", None)
		data = None
		if http_method:
			if http_method == 'GET':
				data = self.event.get('query', None)
			elif http_method == 'POST':
				data = self.event.get('body', None)
			elif http_method == 'PUT':
				data = self.event.get('body').get('data', None)
		if not data and crud_node:
			data = crud_node
		if not data and data_node:
			data = data_node
		if not data and use_event_if_none:
			data = self.event
		if not data and error_on_empty:
			raise ExceptionParamsError("'data' not found in event root neither in 'crud' node")
		return data
	
	def get_filters(self, error_on_empty=True):
		""" get filters node from root or crud node """
		http_method = self.event.get('method', None)
		crud_node = self.event.get("crud", None)
		filter_node = self.event.get("filters", None)
		filters = None
		if http_method:
			if http_method == 'GET':
				filters = self.event.get('query', None)
			elif http_method == 'POST':
				filters = self.event.get('body', None)
		if not filters and crud_node:
			filters = crud_node
		if not filters and filter_node:
			filters = filter_node
		if not filters and error_on_empty:
			raise ExceptionParamsError("'filters' not found in event root neither in 'crud' node")
		_errors = FilterScheme().validate(data=filters)
		if _errors:
			raise Exception("Serializer validation error: {}".format(_errors))
		
		# Filter only required parameters
		if filters:
			return { f:filters[f] for f in FILTER_FIELDS if f in filters }
		else:
			return filters

class ServerlessCrudGet(ServerlessCrudBase):

	def response(self):
		""" We get a element from DB using a primary_key
		
		Parameters:
			1. "primary_keys": it can be inside a "crud" (first search)
			2. "primary_keys": It can be in the event root (second search)
			3. "primary_keys": If 1 and 2 fail, we use the whole received event (json) has primary_keys object
		
		Returns:
			[json] -- Found element of exception
		"""

		try:
			primary_keys = self.get_primary_keys()
			data = self.controller.get(primary_keys=primary_keys)
			return http_response(data=data) if self.json_response else json_decimal_to_float(data=data)
		except ExceptionParamsError as e:
			if self.json_response:
				return http_error(message="primary_keys not found in the event", detail="primary_keys is a dict", e=e)
			else:
				raise e
		except ExceptionInvalidDynamoControllerNotFound as e:
			if self.json_response:
				return http_error(message="Element not found", detail="Not found")
			else:
				raise e
		except Exception as e:
			if self.json_response:
				return http_error(message="Not expected error", detail=str(e))
			else:
				raise e


class ServerlessCrudCreate(ServerlessCrudBase):
	serializer = None

	def __init__(self, event, context, table_name, serializer=None, debug=False):
		self.serializer = serializer if not self.serializer else self.serializer
		self.verify_serializer_defined()
		return super(ServerlessCrudCreate, self).__init__(event=event, context=context, table_name=table_name, debug=debug)

	def response(self):
		try:
			data = self.get_data(use_event_if_none=True)
			self.verify_data_serialized(data=data)
			if self.auto_add_id:
				data["id"] = unique_id()
			data["ts_updated"] = current_ts()
			data_res = self.controller.create(data=data)
			return http_response(data=data_res) if self.json_response else json_decimal_to_float(data=data)
		except ExceptionInvalidDynamoControllerNotFound as e:
			if self.json_response:
				return http_error(message="Element not found", detail="Not found")
			else:
				raise e
		except Exception as e:
			if self.json_response:
				return http_error(message="Not expected error", detail=str(e))
			else:
				raise e

class ServerlessCrudUpdate(ServerlessCrudBase):
	serializer = None

	def __init__(self, event, context, table_name, serializer=None, debug=False):
		self.serializer = serializer if not self.serializer else self.serializer
		self.verify_serializer_defined()
		return super(ServerlessCrudUpdate, self).__init__(event=event, context=context, table_name=table_name, debug=debug)

	def response(self):
		try:
			primary_keys = self.get_primary_keys(use_event_if_none=False, error_on_empty=True)
			data = self.get_data(use_event_if_none=False, error_on_empty=True)
			self.verify_data_serialized(data=data, partial=True)
			data["ts_updated"] = current_ts()
			data_res = self.controller.update(primary_keys=primary_keys, data=data)
			return http_response(data=data_res) if self.json_response else json_decimal_to_float(data=data_res)
		except ExceptionInvalidDynamoControllerNotFound as e:
			if self.json_response:
				return http_error(message="Element not found", detail="Not found")
			else:
				raise e
		except Exception as e:
			if self.json_response:
				return http_error(message="Not expected error", detail=str(e))
			else:
				raise e

class ServerlessCrudDelete(ServerlessCrudBase):
	
	def response(self):
		try:
			primary_keys = self.get_primary_keys()
			data = self.controller.delete(primary_keys=primary_keys)
			return http_response(data=data) if self.json_response else json_decimal_to_float(data=data)
		except ExceptionParamsError as e:
			if self.json_response:
				return http_error(message="primary_keys not found in the event", detail="primary_keys is a dict", e=e)
			else:
				raise e
		except ExceptionInvalidDynamoControllerNotFound as e:
			if self.json_response:
				return http_error(message="Element not found", detail="Not found")
			else:
				raise e
		except Exception as e:
			if self.json_response:
				return http_error(message="Not expected error", detail=str(e))
			else:
				raise e

class ServerlessCrudFilter(ServerlessCrudBase):
	
	def response(self):
		try:
			filters = self.get_filters()
			filtered_data, count, next = self.controller.filter(**filters)
			# TODO: create next_url based on limit and next
			return http_response(data=filtered_data) if self.json_response else json_decimal_to_float(data=filtered_data)
		except Exception as e:
			if self.json_response:
				return http_error(message="Not expected error", detail=str(e))
			else:
				raise e




import os
import sys
import boto3
import logging
import time
from lantern_sl.utils.json import json_float_to_decimal
from lantern_sl.utils.request import http_response, http_error, res_failed
from boto3.dynamodb.conditions import Key, Attr

log = logging.getLogger(__name__)

class ExceptionInvalidDynamoControllerType(Exception):
    """ Dynamo Type Not defined, this is a programming error """
    pass


class ExceptionInvalidDynamoControllerNotFound(Exception):
    """ Object not found in DB (404 error) """
    pass

def get_dynamo_connection(region_name=None, endpoint_url=None, force_live=False, force_local=False):
    """ Return a boto3 Dynamo Connection, live or test if pytest is running cases
    
    Keyword Arguments:
        region_name {str} -- Region to be used for dynamo(default: {None})
        endpoint_url {str} -- endpoint to be used (default: {None})
        force_live {bool} -- Force Live Connection to Dynamo (default: {False})
        force_local {bool} -- Force Local Connection to Dynamo (Default: {False})
    
    Returns:
        [Dynamo Connection] -- Dynamo Connection to be used
    """
    params = {}
    if not force_local:
        if region_name:
            params["region_name"] = region_name
        if endpoint_url:
            params["endpoint_url"] = endpoint_url
    if not force_live:
        if force_local or "pytest" in sys.modules:
            params["endpoint_url"] = "http://localhost:8989"
            params["region_name"] = "us-west-2"
    return boto3.resource("dynamodb", **params)
    

class DynamoController(object):
    """ Generic Dynamo Access Controller
        - Error handling
        - response message/data normalization
    """

    TYPE_GET = "get"
    TYPE_CREATE = "create"
    TYPE_UPDATE = "update"
    TYPE_DELETE = "delete"
    TYPE_FILTER = "filter"

    ORDER_ASC = "asc"
    ORDER_DESC = "desc"

    def __init__(self, table_name, debug=False):
        """ Initialize Dynamo Controller"""
        if not table_name:
            raise Exception("table_name can not be empty")
        self.dynamodb = get_dynamo_connection()
        self.table = self.dynamodb.Table(table_name)
        self.debug = debug if debug else os.getenv("LOCAL_USER", False)

    def get(self, primary_keys):
        """ return a the first element corresponding to the filter_data param"""
        return self._execute_operation(type=self.TYPE_GET, primary_keys=primary_keys)

    def create(self, data):
        """ Creates a new instance in the DB."""
        return self._execute_operation(type=self.TYPE_CREATE, data=data)

    def update(self, primary_keys, data):
        """ Update an existing object in the database"""
        res_update = self._execute_operation(type=self.TYPE_GET, primary_keys=primary_keys)
        if res_failed(res_update):
            return res_update
        self._execute_operation(type=self.TYPE_UPDATE, primary_keys=primary_keys, data=data)
        return self._execute_operation(type=self.TYPE_GET, primary_keys=primary_keys)

    def delete(self, primary_keys):
        """ Deletes a set of objects, corresponding to filter_data"""
        res = self._execute_operation(type=self.TYPE_GET, primary_keys=primary_keys)
        res_del = self._execute_operation(type=self.TYPE_DELETE, primary_keys=primary_keys)
        if res_failed(res_del):
            return res_del
        return res


    def get_filter_fe(self, fe_dict):
        """ generate filter_expression based on fe_dict received in event """
        tot_fe = None
        operation = fe_dict.get('operation', None)
        conditions = fe_dict.get('conditions', None)
        if not operation:
            raise Exception("'operation' is required")
        if not conditions:
            raise Exception("'conditions' is required")
        if operation != 'and' and operation != 'or':
            raise Exception("operation should be 'and' or 'or")
        for f in conditions:
            new_fe = None
            f_type = f.get('type', None)
            param1 = f.get('param1', None)
            param2 = f.get('param2', None)
            param3 = f.get('param3', None)
            if not param1 or not param2:
                raise Exception("type: {} defined but param1 or param2 not".format(f_type))
            if f_type == 'eq':
                new_fe = Attr(param1).eq(param2)
            if f_type == 'ne':
                new_fe = Attr(param1).ne(param2)
            if f_type == 'lt':
                new_fe = Attr(param1).lt(param2)
            if f_type == 'lte':
                new_fe = Attr(param1).lte(param2)
            if f_type == 'gt':
                new_fe = Attr(param1).gt(param2)
            if f_type == 'gte':
                new_fe = Attr(param1).gte(param2)
            if f_type == 'begins_with':
                new_fe = Attr(param1).begins_with(param2)
            if f_type == 'between':
                if not param3:
                    raise Exception("param3 not defined")
                new_fe = Attr(param1).between(param2, param3)

            if new_fe:
                if not tot_fe:
                    tot_fe = new_fe
                else:
                    if operation == "and":
                        tot_fe = tot_fe & new_fe
                    elif operation == "or":
                        tot_fe = tot_fe | new_fe
        return tot_fe
    
    def get_condition_expression_fe(self, c_express):
        """ return condition expression if c_express has all valid keys """
        total_condition = None
        if 'conditions' in c_express and len(c_express):
            for condition in c_express['conditions']:
                op = condition.get('op', None)
                key = condition.get('key', None)
                value = condition.get('value', None)
                value2 = condition.get('value2', None)
                new_fe = None
                if not op or not key or not value:
                    raise Exception("Empty value in get_condition_expression op: {}, key:{}, value: {}, value2: {}".format(op, key, value, value2))
                if op == 'eq':
                    new_fe = Key(key).eq(value)
                if op == 'ne':
                    new_fe = Key(key).ne(value)
                if op == 'lt':
                    new_fe = Key(key).lt(value)
                if op == 'lte':
                    new_fe = Key(key).lte(value)
                if op == 'gt':
                    new_fe = Key(key).gt(value)
                if op == 'gte':
                    new_fe = Key(key).gte(value)
                if op == 'begins_with':
                    new_fe = Key(key).begins_with(value)
                if op == 'between':
                    if not value2:
                        raise Exception("value2 not defined in between, condition_expression")
                    new_fe = Key(key).between(value, value2)
                
                if new_fe:
                    if not total_condition:
                        total_condition = new_fe
                    else:
                        total_condition = total_condition & new_fe
        return total_condition






    def filter(self, key, value, order_by=None, limit=50, next=None, index=None, order_type=ORDER_DESC, filter_expression=None, invert_result=False, condition_expression=None):
        """ Return all orders related to this user
            Params:
                - filter_expression is of type: FilterExpression (boto3), Veryyy flexible
                - first_element: Boolean, if set to True it will return the first element in the query of 404
                - filter_app: Boolean, if set to True, request is requiered and filter_expression will be overrided to filter by app_id
                    - Exception: if no request is provided
                    - ValueError: is filter_expression has a value
        """
        if index:
            index_name = index
        elif key and order_by:
            index_name = "{}-{}-index".format(key, order_by)
        else:
            index_name = "{}-index".format(key)

        params = {
            "IndexName": index_name,
            "Limit":limit,
        }
        if next:
            params["ExclusiveStartKey"] = next


        # Adding conditione expression from parameter or adding default (key/value)
        if condition_expression and type(condition_expression) == dict:
            condition_expression = json_float_to_decimal(condition_expression)
            condition_expression = self.get_condition_expression_fe(condition_expression)
            if condition_expression:
                params["KeyConditionExpression"] = Key(key).eq(value) & condition_expression
        if "KeyConditionExpression" not in params:
            params["KeyConditionExpression"] = Key(key).eq(value)


        #if request:
        #    time_filter,order = self._get_request_parameters(request)
        #    filter_expression = filter_expression & time_filter if time_filter and filterExpression else filter_expression

        if filter_expression and type(filter_expression) == dict:
            filter_expression = json_float_to_decimal(filter_expression)
            filter_expression = self.get_filter_fe(filter_expression)


        if filter_expression:
            params["FilterExpression"] = filter_expression

        #order_type = order if order else order_type
        if order_type == self.ORDER_DESC:
            params["ScanIndexForward"] = False
        data, count, next = self._execute_operation(type=self.TYPE_FILTER, primary_keys=params)
        if invert_result:
            data = list(reversed(data))
        return data, count, next

    '''
    def _get_request_parameters(self,request):
        """ Get order, initial timestamp and final timestamp from request
        """
        order               = request.args.get("order", None)
        initial_timestamp   = request.args.get("initial_timestamp", None)
        final_timestamp     = request.args.get("final_timestamp", None)

        if not final_timestamp and initial_timestamp:
            filter_expression   = Key('timestamp').gt(int(initial_timestamp))

        if not initial_timestamp and final_timestamp:
            filter_expression   = Key('timestamp').lt(int(final_timestamp))

        if not initial_timestamp and not final_timestamp:
            filter_expression   = None

        if initial_timestamp and final_timestamp:
            filter_expression   = Key('timestamp').between(int(initial_timestamp), int(final_timestamp))

        return filter_expression,order
    '''
    
    def _execute_operation(self, type, primary_keys=None, data=None):
        """ Execute the dynamo operation and return a proper response (success or error)"""
        try:
            if type == self.TYPE_GET:
                d_res = self.table.get_item(Key=primary_keys)
                if "Item" in d_res:
                    data = d_res["Item"]
                else:
                    raise ExceptionInvalidDynamoControllerNotFound("Not Found")
                return data
            elif type == self.TYPE_CREATE:
                data = json_float_to_decimal(data)
                d_res = self.table.put_item(Item=data)
                return data
            elif type == self.TYPE_UPDATE:
                data = json_float_to_decimal(data)
                d_res = self.table.update_item(
                    Key=primary_keys,
                    UpdateExpression="set " + ", ".join(["#{} = :_{}".format(key,key) for key in data.keys()]),
                    ExpressionAttributeNames={ "#{}".format(key):key for key in data.keys() },
                    ExpressionAttributeValues={ ":_{}".format(key): data[key] for key in data.keys() },
                    ReturnValues="UPDATED_NEW")
                return data
            elif type == self.TYPE_DELETE:
                d_res = self.table.delete_item(Key=primary_keys)
                return primary_keys
            elif type == self.TYPE_FILTER:
                response = self.table.query(**primary_keys)
                data = response["Items"] if response["Count"] != 0 else []
                code = response['ResponseMetadata']['HTTPStatusCode']
                count = response["Count"]
                next = response["LastEvaluatedKey"] if "LastEvaluatedKey" in response else None
                return data, count, next
            else:
                raise ExceptionInvalidDynamoControllerType("Type {} not defined in DynamoController".format(type))
        except ExceptionInvalidDynamoControllerType as e:
            raise e
        except ExceptionInvalidDynamoControllerNotFound as e:
            raise e
        except Exception as e:
            raise e



"""
Lantern Serverless Utils
"""
from setuptools import find_packages, setup

PACKAGE_VERSION = "0.0.31"

DEPENDENCIES = [
    "boto3==1.7.69",
    "marshmallow==3.0.0rc1"
]

setup(
    name='lantern-sl',
    version=PACKAGE_VERSION,
    url='https://gitlab.com/lantern-tech/lantern-sl.git',
    license='BSD',
    author='Lantern Technologies',
    author_email='support@lantern.tech',
    description='Lantern Serverless Tools',
    long_description=__doc__,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=DEPENDENCIES,
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        #'Programming Language :: Python',
        #'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
